function  [names,exst] = refer(filein,a1,a2,a3,a4)

% REFER Find references to other functions in a file.
%	[NAMES,EXST] = REFER(FILE) Reads text from the
%	file FILE, parses it for all legal names
%	(excluding comments) and returns string matrix
%	NAMES of function names NAMES used in the FILE
%	and vector EXST - their exist status (from 1 to
%	5).
%	Some simple functions and language commands
%	such as 'ones', 'rand', 'sin', 'for' , 'else',
%	etc. are normally excluded. To include them
%	the additional arguments such as 'trivia',
%	'math', 'controls' can be used.
%	See FNAMES.M function for details.
%	REFER(FILE,NEX) where vector NEX with values
%	from 1 to 5 also specifies the exist status
%	of functions to be returned.
%
%	For example:
%	>> refer('interp','math',3:5)
%	 Extracted all names. Checking EXIST status...
%
%	ans =
%
%	error
%	fix
%	sin
%	pi
%	inv
%	filter
%
%	returns only built-in and MEX functions
%	(including all "math" functions) used in the
%	file INTERP.M

%  Kirill K. Pankratov, kirill@plume.mit.edu
%  02/06/95

 % Handle input ..................................
if nargin < 1
  error(' Not enough input arguments');
end

nex = 2:5; % Default exist status

In_str = str2mat('trivia','math','controls');
In_n = [];
for jj = 2:nargin
  eval(['c_arg=a' int2str(jj-1) ';']);

  if isstr(c_arg)  % String - names of matrices
    lm = min(length(c_arg),size(In_str,2));
    a = lower(c_arg(ones(3,1),1:lm))==In_str(1:3,1:lm);
    In_n = [In_n find(sum(a')==lm)];

  else            % Numbers - exist status
    nex = c_arg;
  end
end


 % Read text from a file .........................
e = [exist(filein) exist([filein '.m'])];
if e(2), filein = [filein '.m']; end
string = freadtxt(filein);
string = [' ' string(:)' ' '];  % Add blank edges
lstr = length(string);

 % Compile "trivial" or special names ............
[trivia,math,controls,exim] = fnames;
i0 = ones(1,3);
i0(In_n) = zeros(size(In_n));
M0 = ' ';
for jj = find(i0)
  eval(['M0 = str2mat(M0,' deblank(In_str(jj,:)) ');'])
end
M0 = M0(2:size(M0,1),:);
l_0 = size(M0,1);

 % Auxillary ..................
ch_noas = ('!~=><')';  % Char. before '=' in non-assign.
o_noas = ones(length(ch_noas),1);


 % New lines, strings and comments ...............
[i_nl,mask_q,mask_c] = findquot(string);

 % Blank the openings of quotes (strings) ........
nn = find(diff(mask_q)==1)+1;
string(nn) = ' '*ones(size(nn));

 % Blank the comments
string(mask_c) = ' '*ones(size(find(mask_c)));
string = setstr(string);

string_q = string;
string_q(mask_q) = ' '*ones(size(find(mask_q)));

 % If nothing left, return ......................
if ~any(string>32); return, end

 % Line numbering ...........
num_nl = zeros(1,lstr);
num_nl(i_nl) = ones(1,length(i_nl));
num_nl = cumsum(num_nl)+1;

 % Determine bracketing .....
[brro,brrc,mask_brr] = brackets(string_q);
[brso,brrs,mask_brs] = brackets(string_q,'[');

 % Find end of statements and statements ##
 % (; ,  which are not inside brackets) ...............
i_end = find(string==','|string==';');
i_end = sort([i_nl i_end]);
i_end = i_end( ~mask_brr(i_end) & ~mask_brs(i_end) );
i_end = i_end(diff([0 i_end])>1);

 % Statement numbering ...............
num_st = zeros(1,lstr);
num_st(i_end+1) = ones(size(i_end));
num_st = cumsum(num_st)+1;
n_st = num_st(lstr);  % Total nmb. of statements

 % Find assignment operator for each statement.........
i_eq = find(string=='=');  % All '=' characters
i_eq = i_eq(~mask_c(i_eq) & ~mask_q(i_eq));
l_eq = length(i_eq);
ol = ones(1,l_eq);
 % Separate logical from assignment operators
a = any(string(o_noas,i_eq-1)==ch_noas(:,ol));
a = a | any(string(o_noas,i_eq+1)==ch_noas(:,ol));
if i_eq~=[]
  i_eq = i_eq(~a);  % Indices of assignment operators
end

 % Statement ## for assign. operators ................
i_assg = zeros(1,n_st);
i_assg(num_st(i_eq)) = i_eq;
i_st_assg = find(i_assg); % Statement numbers for each
                      % assgn. operator

 % Extract all variable and function names ...........
[names,no,nc] = getnames(string_q);

 % Determine which side of the assignment operator ...
side = no<i_assg(num_st(no)); % 1-left-hand side, 0-right-hand side

 % Choose only those on the right-hand side
nn = find(~side);
names = names(nn,:);
no = no(nn);
nc = nc(nn);

is_q = mask_q(no);

 % Remove standards: "trivia" such as "zeros", "ones"
 % and "controls" such as "for", "if", "end"
names = str2mat(M0,names);
[names,nr,count] = unique(floor(names)); % Remove repeats
szn = size(names);
 % Remove standards
names = setstr(names(l_0+1:szn(1),:));   % Remove
szn = szn-l_0;


 % Exist status ...............................
msg = ' Extracted all names. Checking EXIST status...\n';
fprintf(msg)

exst = zeros(szn(1),1);
for jj = 1:szn(1)
  name = deblank(names(jj,:));  % Current name
  exst(jj) = exist(name);
  e = 2.1*(exist([name '.mat'])>0);
  exst(jj) = max(exst(jj),e);
end

 % Choose only function or file names .........
nn = zeros(size(exst));
for jj = nex;
  nn = nn | floor(exst)==jj;
end
nn = find(nn);
names = names(nn,:);
exst = exst(nn);
